from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import BinaryIO, Dict, Iterator, Optional, Tuple, TypeVar, Union

import torch
from torch.utils._pytree import tree_map

from humecodec._extension import get_ext


@dataclass
class SourceStream:
    """The metadata of a source stream."""

    media_type: str
    """The type of the stream."""
    codec: str
    """Short name of the codec."""
    codec_long_name: str
    """Detailed name of the codec."""
    format: Optional[str]
    """Media format."""
    bit_rate: Optional[int]
    """Bit rate of the stream in bits-per-second."""
    num_frames: Optional[int]
    """The number of frames in the stream."""
    bits_per_sample: Optional[int]
    """The number of valid bits in each output sample."""
    metadata: Dict[str, str]
    """Metadata attached to the source stream."""
    time_base: Tuple[int, int]
    """Stream time base as (numerator, denominator). PTS values are in these units."""


@dataclass
class SourceAudioStream(SourceStream):
    """The metadata of an audio source stream."""

    sample_rate: float
    """Sample rate of the audio."""
    num_channels: int
    """Number of channels."""


@dataclass
class SourceVideoStream(SourceStream):
    """The metadata of a video source stream."""

    width: int
    """Width of the video frame in pixel."""
    height: int
    """Height of the video frame in pixel."""
    frame_rate: float
    """Frame rate."""


def _parse_si(i):
    media_type = i.media_type
    tb = (i.time_base_num, i.time_base_den)
    if media_type == "audio":
        return SourceAudioStream(
            media_type=i.media_type,
            codec=i.codec_name,
            codec_long_name=i.codec_long_name,
            format=i.format,
            bit_rate=i.bit_rate,
            num_frames=i.num_frames,
            bits_per_sample=i.bits_per_sample,
            metadata=i.metadata,
            time_base=tb,
            sample_rate=i.sample_rate,
            num_channels=i.num_channels,
        )
    if media_type == "video":
        return SourceVideoStream(
            media_type=i.media_type,
            codec=i.codec_name,
            codec_long_name=i.codec_long_name,
            format=i.format,
            bit_rate=i.bit_rate,
            num_frames=i.num_frames,
            bits_per_sample=i.bits_per_sample,
            metadata=i.metadata,
            time_base=tb,
            width=i.width,
            height=i.height,
            frame_rate=i.frame_rate,
        )
    return SourceStream(
        media_type=i.media_type,
        codec=i.codec_name,
        codec_long_name=i.codec_long_name,
        format=None,
        bit_rate=None,
        num_frames=None,
        bits_per_sample=None,
        metadata=i.metadata,
        time_base=tb,
    )


@dataclass
class OutputStream:
    """Output stream configured on MediaDecoder."""

    source_index: int
    """Index of the source stream that this output stream is connected."""
    filter_description: str
    """Description of filter graph applied to the source stream."""
    media_type: str
    """The type of the stream. ``"audio"`` or ``"video"``."""
    format: str
    """Media format."""


@dataclass
class OutputAudioStream(OutputStream):
    """Information about an audio output stream."""

    sample_rate: float
    """Sample rate of the audio."""
    num_channels: int
    """Number of channels."""


@dataclass
class OutputVideoStream(OutputStream):
    """Information about a video output stream."""

    width: int
    """Width of the video frame in pixel."""
    height: int
    """Height of the video frame in pixel."""
    frame_rate: float
    """Frame rate."""


def _parse_oi(i):
    media_type = i.media_type
    if media_type == "audio":
        return OutputAudioStream(
            source_index=i.source_index,
            filter_description=i.filter_description,
            media_type=i.media_type,
            format=i.format,
            sample_rate=i.sample_rate,
            num_channels=i.num_channels,
        )
    if media_type == "video":
        return OutputVideoStream(
            source_index=i.source_index,
            filter_description=i.filter_description,
            media_type=i.media_type,
            format=i.format,
            width=i.width,
            height=i.height,
            frame_rate=i.frame_rate,
        )
    raise ValueError(f"Unexpected media_type: {i.media_type}({i})")


def _get_afilter_desc(sample_rate, fmt, num_channels):
    descs = []
    if sample_rate is not None:
        descs.append(f"aresample={sample_rate}")
    if fmt is not None or num_channels is not None:
        parts = []
        if fmt is not None:
            parts.append(f"sample_fmts={fmt}")
        if num_channels is not None:
            parts.append(f"channel_layouts={num_channels}c")
        descs.append(f"aformat={':'.join(parts)}")
    return ",".join(descs) if descs else None


def _get_vfilter_desc(frame_rate, width, height, fmt):
    descs = []
    if frame_rate is not None:
        descs.append(f"fps={frame_rate}")
    scales = []
    if width is not None:
        scales.append(f"width={width}")
    if height is not None:
        scales.append(f"height={height}")
    if scales:
        descs.append(f"scale={':'.join(scales)}")
    if fmt is not None:
        descs.append(f"format=pix_fmts={fmt}")
    return ",".join(descs) if descs else None


# Interlaced pixel formats that C++ outputs as NHWC (need permute to NCHW)
_NHWC_VIDEO_FORMATS = {
    "rgb24", "bgr24", "argb", "rgba", "abgr", "bgra", "gray8", "rgb48le",
}


class ChunkTensorBase(torch.Tensor):
    __torch_function__ = torch._C._disabled_torch_function_impl

    @staticmethod
    def __new__(cls, _elem, *_):
        return super().__new__(cls, _elem)

    @classmethod
    def __torch_dispatch__(cls, func, _, args=(), kwargs=None):
        def unwrap(t):
            return t._elem if isinstance(t, cls) else t

        return func(*tree_map(unwrap, args), **tree_map(unwrap, kwargs))


@dataclass
class Chunk(ChunkTensorBase):
    """Decoded media frames with metadata.

    The instance of this class represents the decoded video/audio frames with
    metadata, and the instance itself behaves like :py:class:`~torch.Tensor`.
    """

    _elem: torch.Tensor

    pts: float
    """Presentation time stamp of the first frame in the chunk. Unit: second."""


InputStreamTypes = TypeVar("InputStream", bound=SourceStream)
OutputStreamTypes = TypeVar("OutputStream", bound=OutputStream)


class MediaDecoder:
    """Fetch and decode audio/video streams chunk by chunk.

    Args:
        src (str, path-like, bytes or file-like object): The media source.
        format (str or None, optional): Override the input format.
            Default: ``None``.
        option (dict of str to str, optional): Custom option passed when
            initializing format context. Default: ``None``.
        buffer_size (int): The internal buffer size in byte. Used only when
            `src` is file-like object. Default: ``4096``.
    """

    def __init__(
        self,
        src: Union[str, Path, BinaryIO],
        format: Optional[str] = None,
        option: Optional[Dict[str, str]] = None,
        buffer_size: int = 4096,
    ):
        ext = get_ext()
        self.src = src
        if isinstance(src, bytes):
            self._be = ext.StreamingMediaDecoderBytes(src, format, option, buffer_size)
        elif hasattr(src, "read"):
            self._be = ext.StreamingMediaDecoderFileObj(src, format, option, buffer_size)
        else:
            self._be = ext.StreamingMediaDecoder(os.path.normpath(src), format, option)

        i = self._be.find_best_audio_stream()
        self._default_audio_stream = None if i < 0 else i
        i = self._be.find_best_video_stream()
        self._default_video_stream = None if i < 0 else i
        self._out_transforms: Optional[list] = None

    @property
    def num_src_streams(self):
        """Number of streams found in the provided media source."""
        return self._be.num_src_streams()

    @property
    def num_out_streams(self):
        """Number of output streams configured by client code."""
        return self._be.num_out_streams()

    @property
    def default_audio_stream(self):
        """The index of default audio stream. ``None`` if there is no audio stream."""
        return self._default_audio_stream

    @property
    def default_video_stream(self):
        """The index of default video stream. ``None`` if there is no video stream."""
        return self._default_video_stream

    def get_metadata(self) -> Dict[str, str]:
        """Get the metadata of the source media."""
        return self._be.get_metadata()

    def get_src_stream_info(self, i: int) -> InputStreamTypes:
        """Get the metadata of source stream."""
        return _parse_si(self._be.get_src_stream_info(i))

    def get_out_stream_info(self, i: int) -> OutputStreamTypes:
        """Get the metadata of output stream."""
        info = self._be.get_out_stream_info(i)
        return _parse_oi(info)

    def seek(self, timestamp: float, mode: str = "precise"):
        """Seek the stream to the given timestamp [second].

        Args:
            timestamp (float): Target time in second.
            mode (str): Controls how seek is done.
                Valid choices are ``"key"``, ``"any"``, ``"precise"``.
        """
        modes = {
            "key": 0,
            "any": 1,
            "precise": 2,
        }
        if mode not in modes:
            raise ValueError(f"The value of mode must be one of {list(modes.keys())}. Found: {mode}")
        self._be.seek(timestamp, modes[mode])

    def build_packet_index(self, stream_index: Optional[int] = None, resolution: int = 524288):
        """Scan all packets for a stream and return a lightweight seek index.

        Each entry contains ``pts_seconds``, ``pos`` (byte offset), ``size``,
        and ``is_key``.  The byte offsets can be passed to
        :meth:`seek_to_byte_offset` for instantaneous seeking without any
        container-level scanning.

        After this call the read position is at EOF; call :meth:`seek` or
        :meth:`seek_to_byte_offset` before continuing to decode.

        Args:
            stream_index (int or None, optional): Source stream to index.
                Defaults to the best audio stream.
            resolution (int, optional): Minimum byte distance between
                consecutive index entries.  Defaults to 524288 (512 KB).
                Use ``resolution=1`` for a full per-packet index.
        """
        if stream_index is None:
            stream_index = self._default_audio_stream
        if stream_index is None:
            raise RuntimeError("No stream to index (no audio stream found and no stream_index given).")
        return self._be.build_packet_index(stream_index, resolution)

    def seek_to_byte_offset(self, offset: int):
        """Seek directly to a byte position in the underlying file.

        This is intended for use with offsets obtained from
        :meth:`build_packet_index` and skips all container-level parsing.

        Args:
            offset (int): Byte offset (from :attr:`PacketIndexEntry.pos`).
        """
        self._be.seek_to_byte_offset(offset)

    def add_basic_audio_stream(
        self,
        frames_per_chunk: int,
        buffer_chunk_size: int = 3,
        *,
        stream_index: Optional[int] = None,
        decoder: Optional[str] = None,
        decoder_option: Optional[Dict[str, str]] = None,
        format: Optional[str] = "fltp",
        sample_rate: Optional[int] = None,
        num_channels: Optional[int] = None,
    ):
        """Add output audio stream with basic options."""
        self.add_audio_stream(
            frames_per_chunk,
            buffer_chunk_size,
            stream_index=stream_index,
            decoder=decoder,
            decoder_option=decoder_option,
            filter_desc=_get_afilter_desc(sample_rate, format, num_channels),
        )

    def add_basic_video_stream(
        self,
        frames_per_chunk: int,
        buffer_chunk_size: int = 3,
        *,
        stream_index: Optional[int] = None,
        decoder: Optional[str] = None,
        decoder_option: Optional[Dict[str, str]] = None,
        format: Optional[str] = "rgb24",
        frame_rate: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        hw_accel: Optional[str] = None,
    ):
        """Add output video stream with basic options."""
        self.add_video_stream(
            frames_per_chunk,
            buffer_chunk_size,
            stream_index=stream_index,
            decoder=decoder,
            decoder_option=decoder_option,
            filter_desc=_get_vfilter_desc(frame_rate, width, height, format),
            hw_accel=hw_accel,
        )

    def add_audio_stream(
        self,
        frames_per_chunk: int,
        buffer_chunk_size: int = 3,
        *,
        stream_index: Optional[int] = None,
        decoder: Optional[str] = None,
        decoder_option: Optional[Dict[str, str]] = None,
        filter_desc: Optional[str] = None,
    ):
        """Add output audio stream.

        Args:
            frames_per_chunk (int): Number of frames returned as one chunk.
            buffer_chunk_size (int, optional): Internal buffer size. Default: ``3``.
            stream_index (int or None, optional): The source audio stream index.
            decoder (str or None, optional): The name of the decoder.
            decoder_option (dict or None, optional): Options passed to decoder.
            filter_desc (str or None, optional): Filter description.
        """
        i = self.default_audio_stream if stream_index is None else stream_index
        if i is None:
            raise RuntimeError("There is no audio stream.")
        self._out_transforms = None
        self._be.add_audio_stream(
            i,
            frames_per_chunk,
            buffer_chunk_size,
            filter_desc,
            decoder,
            decoder_option or {},
        )

    def add_video_stream(
        self,
        frames_per_chunk: int,
        buffer_chunk_size: int = 3,
        *,
        stream_index: Optional[int] = None,
        decoder: Optional[str] = None,
        decoder_option: Optional[Dict[str, str]] = None,
        filter_desc: Optional[str] = None,
        hw_accel: Optional[str] = None,
    ):
        """Add output video stream.

        Args:
            frames_per_chunk (int): Number of frames returned as one chunk.
            buffer_chunk_size (int, optional): Internal buffer size. Default: ``3``.
            stream_index (int or None, optional): The source video stream index.
            decoder (str or None, optional): The name of the decoder.
            decoder_option (dict or None, optional): Options passed to decoder.
            filter_desc (str or None, optional): Filter description.
            hw_accel (str or None, optional): Enable hardware acceleration.
        """
        i = self.default_video_stream if stream_index is None else stream_index
        if i is None:
            raise RuntimeError("There is no video stream.")
        self._out_transforms = None
        self._be.add_video_stream(
            i,
            frames_per_chunk,
            buffer_chunk_size,
            filter_desc,
            decoder,
            decoder_option or {},
            hw_accel,
        )

    def remove_stream(self, i: int):
        """Remove an output stream."""
        self._out_transforms = None
        self._be.remove_stream(i)

    def process_packet(self, timeout: Optional[float] = None, backoff: float = 10.0) -> int:
        """Read the source media and process one packet.

        Returns:
            int: ``0`` if a packet was processed, ``1`` if EOF was reached.
        """
        return self._be.process_packet(timeout, backoff)

    def process_all_packets(self):
        """Process packets until EOF."""
        self._be.process_all_packets()

    def is_buffer_ready(self) -> bool:
        """Returns true if all output streams have at least one chunk filled."""
        return self._be.is_buffer_ready()

    def _ensure_transforms(self):
        """Build per-output-stream transform list lazily."""
        n = self._be.num_out_streams()
        if self._out_transforms is not None and len(self._out_transforms) == n:
            return
        transforms = []
        for idx in range(n):
            info = self._be.get_out_stream_info(idx)
            if info.media_type == "video" and info.format in _NHWC_VIDEO_FORMATS:
                transforms.append("nhwc_to_nchw")
            else:
                transforms.append(None)
        self._out_transforms = transforms

    def pop_chunks(self) -> Tuple[Optional[Chunk]]:
        """Pop one chunk from all the output stream buffers.

        Returns decoded frames as Chunk objects, which wrap torch.Tensor instances.
        The tensors are transferred via DLPack protocol for efficient zero-copy exchange.
        """
        self._ensure_transforms()
        ret = []
        for idx, item in enumerate(self._be.pop_chunks_dlpack()):
            if item is None:
                ret.append(None)
            else:
                capsule, pts = item
                tensor = torch.from_dlpack(capsule)
                if self._out_transforms[idx] == "nhwc_to_nchw":
                    tensor = tensor.permute(0, 3, 1, 2).contiguous()
                ret.append(Chunk(tensor, pts))
        return ret

    def fill_buffer(self, timeout: Optional[float] = None, backoff: float = 10.0) -> int:
        """Keep processing packets until all buffers have at least one chunk.

        Returns:
            int: ``0`` if buffers are ready, ``1`` if EOF was reached.
        """
        return self._be.fill_buffer(timeout, backoff)

    def stream(
        self, timeout: Optional[float] = None, backoff: float = 10.0
    ) -> Iterator[Tuple[Optional[Chunk], ...]]:
        """Return an iterator that generates output tensors."""
        if self.num_out_streams == 0:
            raise RuntimeError("No output stream is configured.")

        while True:
            if self.fill_buffer(timeout, backoff):
                break
            yield self.pop_chunks()

        while True:
            chunks = self.pop_chunks()
            if all(c is None for c in chunks):
                return
            yield chunks



"""Round-trip encode/decode tests for humecodec."""

import math
import torch
import pytest
import humecodec


def _make_sine(sample_rate: int = 44100, duration: float = 1.0,
               freq: float = 440.0, num_channels: int = 1) -> torch.Tensor:
    """Generate a sine-wave tensor of shape (num_frames, num_channels)."""
    n = int(sample_rate * duration)
    t = torch.arange(n, dtype=torch.float32) / sample_rate
    wave = 0.5 * torch.sin(2 * math.pi * freq * t)
    return wave.unsqueeze(1).expand(-1, num_channels).contiguous()


def test_save_load_wav_roundtrip(tmp_path):
    path = str(tmp_path / "test.wav")
    sr = 44100
    original = _make_sine(sample_rate=sr, duration=0.5)

    humecodec.save_audio(path, original, sr)
    loaded, loaded_sr = humecodec.load_audio(path)

    assert loaded_sr == sr
    # Codec may pad frames; compare the overlapping region
    min_frames = min(original.shape[0], loaded.shape[0])
    assert loaded.shape[1] == original.shape[1]
    torch.testing.assert_close(loaded[:min_frames], original[:min_frames], atol=1e-4, rtol=1e-4)


def test_save_load_flac_roundtrip(tmp_path):
    path = str(tmp_path / "test.flac")
    sr = 44100
    original = _make_sine(sample_rate=sr, duration=0.5)

    humecodec.save_audio(path, original, sr)
    loaded, loaded_sr = humecodec.load_audio(path)

    assert loaded_sr == sr
    # FLAC may pad to block boundaries; compare the overlapping region
    min_frames = min(original.shape[0], loaded.shape[0])
    assert loaded.shape[1] == original.shape[1]
    torch.testing.assert_close(loaded[:min_frames], original[:min_frames], atol=1e-4, rtol=1e-4)


def test_save_load_mp3_roundtrip(tmp_path):
    path = str(tmp_path / "test.mp3")
    sr = 44100
    original = _make_sine(sample_rate=sr, duration=1.0)

    humecodec.save_audio(path, original, sr)
    loaded, loaded_sr = humecodec.load_audio(path)

    assert loaded_sr == sr
    # MP3 pads frames, so loaded may be slightly longer; just check it's close
    min_frames = min(original.shape[0], loaded.shape[0])
    assert min_frames > 0
    assert loaded.shape[1] == original.shape[1]
    # Lossy: allow generous tolerance
    diff = (loaded[:min_frames] - original[:min_frames]).abs().mean()
    assert diff < 0.05, f"Mean absolute difference too large: {diff}"


def test_save_load_opus_roundtrip(tmp_path):
    path = str(tmp_path / "test.opus")
    sr = 48000  # Opus native rate
    original = _make_sine(sample_rate=sr, duration=1.0)

    humecodec.save_audio(path, original, sr)
    loaded, loaded_sr = humecodec.load_audio(path)

    assert loaded_sr == sr
    min_frames = min(original.shape[0], loaded.shape[0])
    assert min_frames > 0
    assert loaded.shape[1] == original.shape[1]
    diff = (loaded[:min_frames] - original[:min_frames]).abs().mean()
    assert diff < 0.05, f"Mean absolute difference too large: {diff}"


def test_audio_info(tmp_path):
    path = str(tmp_path / "info_test.wav")
    sr = 16000
    num_channels = 2
    duration = 0.5
    original = _make_sine(sample_rate=sr, duration=duration, num_channels=num_channels)

    humecodec.save_audio(path, original, sr)
    ai = humecodec.info(path)

    assert ai.sample_rate == sr
    assert ai.num_channels == num_channels
    # num_frames may be 0 if the container doesn't store duration in the header
    if ai.num_frames > 0:
        assert ai.num_frames == original.shape[0]


def test_encoder_decoder_streaming(tmp_path):
    path = str(tmp_path / "stream.wav")
    sr = 16000
    num_channels = 1
    chunk_frames = 1024
    num_chunks = 4

    # Encode chunks
    enc = humecodec.MediaEncoder(path)
    enc.add_audio_stream(sample_rate=sr, num_channels=num_channels)
    enc.open()
    for i in range(num_chunks):
        chunk = _make_sine(sample_rate=sr,
                           duration=chunk_frames / sr,
                           freq=440.0,
                           num_channels=num_channels)
        enc.write_audio_chunk(0, chunk)
    enc.flush()
    enc.close()

    # Decode back
    decoder = humecodec.MediaDecoder(path)
    decoder.add_audio_stream(frames_per_chunk=-1)
    decoder.process_all_packets()
    (result,) = decoder.pop_chunks()

    expected_frames = chunk_frames * num_chunks
    assert result is not None
    # Codec may pad; just verify we got at least the expected number of frames
    assert result.shape[0] >= expected_frames
    assert result.shape[1] == num_channels


def test_build_packet_index(tmp_path):
    path = str(tmp_path / "index_test.wav")
    sr = 44100
    original = _make_sine(sample_rate=sr, duration=5.0)
    humecodec.save_audio(path, original, sr)

    # Build index with default resolution
    decoder = humecodec.MediaDecoder(path)
    decoder.add_audio_stream(frames_per_chunk=-1)
    index_default = decoder.build_packet_index()

    assert len(index_default) > 0
    for entry in index_default:
        assert entry.pos >= 0
        assert entry.size > 0
        assert entry.pts_seconds >= 0
    # Entries should be sorted by pts_seconds
    pts_list = [e.pts_seconds for e in index_default]
    assert pts_list == sorted(pts_list)

    # Build index with resolution=1 (every packet)
    decoder2 = humecodec.MediaDecoder(path)
    decoder2.add_audio_stream(frames_per_chunk=-1)
    index_full = decoder2.build_packet_index(resolution=1)

    assert len(index_full) >= len(index_default)


def test_seek_to_byte_offset_sample_accurate(tmp_path):
    path = str(tmp_path / "sweep.wav")
    sr = 44100
    duration = 5.0
    n = int(sr * duration)
    # Frequency sweep so each position is unique
    t = torch.arange(n, dtype=torch.float32) / sr
    freq = 200 + 2000 * t / duration  # 200 Hz -> 2200 Hz
    phase = 2 * math.pi * torch.cumsum(freq / sr, dim=0)
    waveform = (0.5 * torch.sin(phase)).unsqueeze(1)

    humecodec.save_audio(path, waveform, sr)

    # Load the full file as reference (WAV may pad slightly)
    ref, ref_sr = humecodec.load_audio(path)
    assert ref_sr == sr

    # Build full packet index
    decoder = humecodec.MediaDecoder(path)
    decoder.add_audio_stream(frames_per_chunk=-1)
    index = decoder.build_packet_index(resolution=1)
    assert len(index) > 1

    # Pick several entries spread across the file
    step = max(1, len(index) // 5)
    test_entries = index[::step]

    for entry in test_entries:
        dec = humecodec.MediaDecoder(path)
        dec.add_audio_stream(frames_per_chunk=-1, filter_desc="aformat=sample_fmts=fltp")
        dec.seek_to_byte_offset(entry.pos)
        dec.process_all_packets()
        (chunk,) = dec.pop_chunks()
        assert chunk is not None
        assert chunk.shape[0] > 0

        # Determine the sample offset from pts_seconds
        sample_offset = int(round(entry.pts_seconds * sr))
        compare_len = min(chunk.shape[0], ref.shape[0] - sample_offset)
        if compare_len <= 0:
            continue
        ref_region = ref[sample_offset : sample_offset + compare_len]
        torch.testing.assert_close(
            chunk[:compare_len], ref_region, atol=1e-4, rtol=1e-4
        )


def test_get_versions():
    versions = humecodec.get_versions()
    # libavdevice is loaded at runtime via dlopen and may not be available
    expected_keys = {"libavutil", "libavcodec", "libavformat",
                     "libavfilter"}
    assert expected_keys.issubset(set(versions.keys()))
    for name, ver in versions.items():
        assert len(ver) == 3, f"{name} version should be a 3-tuple"
        assert all(isinstance(v, int) for v in ver)



"""humecodec - Standalone FFmpeg integration for PyTorch.

Provides MediaDecoder/MediaEncoder classes and convenience functions
for loading/saving audio and video with PyTorch tensors.
"""

from __future__ import annotations

from typing import Dict, Optional, Tuple

import torch

from humecodec._extension import get_ext
from humecodec.decoder import (
    Chunk,
    MediaDecoder,
    OutputAudioStream,
    OutputStream,
    OutputVideoStream,
    SourceAudioStream,
    SourceStream,
    SourceVideoStream,
)
from humecodec.encoder import CodecConfig, MediaEncoder

__all__ = [
    "MediaDecoder",
    "MediaEncoder",
    "CodecConfig",
    "Chunk",
    "SourceStream",
    "SourceAudioStream",
    "SourceVideoStream",
    "OutputStream",
    "OutputAudioStream",
    "OutputVideoStream",
    "load_audio",
    "save_audio",
    "info",
    "get_audio_decoders",
    "get_audio_encoders",
    "get_video_decoders",
    "get_video_encoders",
    "get_muxers",
    "get_demuxers",
    "get_versions",
]


class AudioInfo:
    """Metadata about an audio file."""

    def __init__(self, sample_rate: int, num_channels: int, num_frames: int, codec: str, format: str):
        self.sample_rate = sample_rate
        self.num_channels = num_channels
        self.num_frames = num_frames
        self.codec = codec
        self.format = format

    def __repr__(self):
        return (
            f"AudioInfo(sample_rate={self.sample_rate}, "
            f"num_channels={self.num_channels}, "
            f"num_frames={self.num_frames}, "
            f"codec='{self.codec}', format='{self.format}')"
        )


def load_audio(
    path: str,
    format: Optional[str] = None,
    offset: float = 0.0,
    duration: Optional[float] = None,
    sample_rate: Optional[int] = None,
    num_channels: Optional[int] = None,
) -> Tuple[torch.Tensor, int]:
    """Load an audio file.

    Args:
        path: Path to the audio file.
        format: Override the input format. Default: ``None``.
        offset: Start reading from this time (seconds). Default: ``0.0``.
        duration: Maximum duration to read (seconds). Default: ``None`` (read all).
        sample_rate: If provided, resample to this rate.
        num_channels: If provided, remix to this many channels.

    Returns:
        Tuple of (waveform, sample_rate) where waveform has shape ``(num_frames, num_channels)``
        and dtype ``torch.float32``.
    """
    decoder = MediaDecoder(path, format=format)

    # Build filter description for resampling/remixing
    filter_parts = []
    if sample_rate is not None:
        filter_parts.append(f"aresample={sample_rate}")
    filter_parts.append("aformat=sample_fmts=fltp")
    if num_channels is not None:
        filter_parts[-1] += f":channel_layouts={num_channels}c"
    filter_desc = ",".join(filter_parts) if filter_parts else None

    decoder.add_audio_stream(
        frames_per_chunk=-1,
        buffer_chunk_size=-1,
        filter_desc=filter_desc,
    )

    if offset > 0:
        decoder.seek(offset)

    decoder.process_all_packets()

    chunks = decoder.pop_chunks()
    if chunks[0] is None:
        raise RuntimeError(f"Failed to decode audio from '{path}'")

    waveform = chunks[0]._elem if hasattr(chunks[0], "_elem") else torch.Tensor(chunks[0])

    if duration is not None:
        # Determine effective sample rate
        out_info = decoder.get_out_stream_info(0)
        sr = int(out_info.sample_rate)
        max_frames = int(duration * sr)
        if waveform.size(0) > max_frames:
            waveform = waveform[:max_frames]

    # Get the actual output sample rate
    out_info = decoder.get_out_stream_info(0)
    sr = int(out_info.sample_rate)

    return waveform, sr


def save_audio(
    path: str,
    waveform: torch.Tensor,
    sample_rate: int,
    format: Optional[str] = None,
    encoder: Optional[str] = None,
    encoder_option: Optional[Dict[str, str]] = None,
    codec_config: Optional[CodecConfig] = None,
):
    """Save a waveform tensor to an audio file.

    Args:
        path: Destination path.
        waveform: Audio tensor. Shape: ``(num_frames, num_channels)``.
        sample_rate: Sample rate of the waveform.
        format: Override the output format. Default: ``None``.
        encoder: The encoder name. Default: ``None``.
        encoder_option: Options for encoder. Default: ``None``.
        codec_config: Codec configuration. Default: ``None``.
    """
    if waveform.dim() != 2:
        raise ValueError(f"Expected 2D tensor (frames, channels), got {waveform.dim()}D")

    num_channels = waveform.size(1)

    enc = MediaEncoder(path, format=format)
    enc.add_audio_stream(
        sample_rate=sample_rate,
        num_channels=num_channels,
        format="flt",
        encoder=encoder,
        encoder_option=encoder_option,
        codec_config=codec_config,
    )

    with enc.open():
        enc.write_audio_chunk(0, waveform.float().contiguous())


def info(path: str, format: Optional[str] = None) -> AudioInfo:
    """Get audio metadata without decoding.

    Args:
        path: Path to the audio file.
        format: Override the input format. Default: ``None``.

    Returns:
        AudioInfo with sample_rate, num_channels, num_frames, codec, format.
    """
    decoder = MediaDecoder(path, format=format)
    if decoder.default_audio_stream is None:
        raise RuntimeError(f"No audio stream found in '{path}'")

    si = decoder.get_src_stream_info(decoder.default_audio_stream)
    return AudioInfo(
        sample_rate=int(si.sample_rate),
        num_channels=si.num_channels,
        num_frames=si.num_frames if si.num_frames else 0,
        codec=si.codec,
        format=si.format or "",
    )


# FFmpeg query functions
def get_audio_decoders() -> Dict[str, str]:
    """List available audio decoders."""
    return get_ext().get_audio_decoders()


def get_audio_encoders() -> Dict[str, str]:
    """List available audio encoders."""
    return get_ext().get_audio_encoders()


def get_video_decoders() -> Dict[str, str]:
    """List available video decoders."""
    return get_ext().get_video_decoders()


def get_video_encoders() -> Dict[str, str]:
    """List available video encoders."""
    return get_ext().get_video_encoders()


def get_muxers() -> Dict[str, str]:
    """List available muxers (output formats)."""
    return get_ext().get_muxers()


def get_demuxers() -> Dict[str, str]:
    """List available demuxers (input formats)."""
    return get_ext().get_demuxers()


def get_versions() -> Dict[str, Tuple[int, int, int]]:
    """Get FFmpeg library versions."""
    return get_ext().get_versions()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import BinaryIO, Dict, Optional, Union

import torch

from humecodec._extension import get_ext


@dataclass
class CodecConfig:
    """Codec configuration."""

    bit_rate: int = -1
    """Bit rate."""

    compression_level: int = -1
    """Compression level."""

    qscale: Optional[int] = None
    """Global quality factor. Enables variable bit rate."""

    gop_size: int = -1
    """The number of pictures in a group of pictures, or 0 for intra_only."""

    max_b_frames: int = -1
    """Maximum number of B-frames between non-B-frames."""


def _convert_config(cfg: Optional[CodecConfig]):
    if cfg is None:
        return None
    ext = get_ext()
    return ext.CodecConfig(
        cfg.bit_rate,
        cfg.compression_level,
        cfg.qscale,
        cfg.gop_size,
        cfg.max_b_frames,
    )


_INTERLACED_FORMATS = {"rgb24", "bgr24", "gray8"}
_INTERLACED_4CH_FORMATS = {"rgb0", "bgr0"}
_PLANAR_FORMATS = {"yuv444p", "gbrp"}


def _prepare_video_tensor(chunk: torch.Tensor, fmt: str) -> torch.Tensor:
    """Prepare a video tensor for encoding by converting to the layout C++ expects.

    C++ expects:
    - Interlaced formats: NHWC contiguous
    - Planar formats: NCHW contiguous
    """
    if fmt in _INTERLACED_FORMATS:
        # NCHW -> NHWC
        return chunk.permute(0, 2, 3, 1).contiguous()
    if fmt in _INTERLACED_4CH_FORMATS:
        if chunk.size(1) == 3:
            # Pad 3-channel NCHW to 4-channel NHWC (padding channel = 0)
            padding = torch.zeros(
                chunk.size(0), 1, chunk.size(2), chunk.size(3),
                dtype=chunk.dtype, device=chunk.device,
            )
            chunk = torch.cat([chunk, padding], dim=1)
        # NCHW -> NHWC (now 4 channels)
        return chunk.permute(0, 2, 3, 1).contiguous()
    if fmt in _PLANAR_FORMATS:
        return chunk.contiguous()
    # Unknown format — pass through contiguous, let C++ handle it
    return chunk.contiguous()


class MediaEncoder:
    """Encode and write audio/video streams chunk by chunk.

    Args:
        dst (str, path-like or file-like object): The destination where
            encoded data are written.
        format (str or None, optional): Override the output format.
            Default: ``None``.
        buffer_size (int): The internal buffer size in byte. Used only when
            `dst` is a file-like object. Default: ``4096``.
    """

    def __init__(
        self,
        dst: Union[str, Path, BinaryIO],
        format: Optional[str] = None,
        buffer_size: int = 4096,
    ):
        ext = get_ext()
        if hasattr(dst, "write"):
            self._s = ext.StreamingMediaEncoderFileObj(dst, format, buffer_size)
        else:
            self._s = ext.StreamingMediaEncoder(str(dst), format)
        self._is_open = False
        self._stream_configs: list = []

    def add_audio_stream(
        self,
        sample_rate: int,
        num_channels: int,
        format: str = "flt",
        *,
        encoder: Optional[str] = None,
        encoder_option: Optional[Dict[str, str]] = None,
        encoder_sample_rate: Optional[int] = None,
        encoder_num_channels: Optional[int] = None,
        encoder_format: Optional[str] = None,
        codec_config: Optional[CodecConfig] = None,
        filter_desc: Optional[str] = None,
    ):
        """Add an output audio stream.

        Args:
            sample_rate (int): The sample rate.
            num_channels (int): The number of channels.
            format (str, optional): Input sample format. Default: ``"flt"``.
            encoder (str or None, optional): The encoder name. Default: ``None``.
            encoder_option (dict or None, optional): Options for encoder.
            encoder_sample_rate (int or None, optional): Override encoding sample rate.
            encoder_num_channels (int or None, optional): Override encoding channels.
            encoder_format (str or None, optional): Override encoding format.
            codec_config (CodecConfig or None, optional): Codec configuration.
            filter_desc (str or None, optional): Additional filter processing.
        """
        self._stream_configs.append(("audio", format))
        self._s.add_audio_stream(
            sample_rate,
            num_channels,
            format,
            encoder,
            encoder_option,
            encoder_format,
            encoder_sample_rate,
            encoder_num_channels,
            _convert_config(codec_config),
            filter_desc,
        )

    def add_video_stream(
        self,
        frame_rate: float,
        width: int,
        height: int,
        format: str = "rgb24",
        *,
        encoder: Optional[str] = None,
        encoder_option: Optional[Dict[str, str]] = None,
        encoder_frame_rate: Optional[float] = None,
        encoder_width: Optional[int] = None,
        encoder_height: Optional[int] = None,
        encoder_format: Optional[str] = None,
        codec_config: Optional[CodecConfig] = None,
        filter_desc: Optional[str] = None,
        hw_accel: Optional[str] = None,
    ):
        """Add an output video stream.

        Args:
            frame_rate (float): Frame rate of the video.
            width (int): Width of the video frame.
            height (int): Height of the video frame.
            format (str, optional): Input pixel format. Default: ``"rgb24"``.
            encoder (str or None, optional): The encoder name.
            encoder_option (dict or None, optional): Options for encoder.
            encoder_frame_rate (float or None, optional): Override encoding frame rate.
            encoder_width (int or None, optional): Override encoding width.
            encoder_height (int or None, optional): Override encoding height.
            encoder_format (str or None, optional): Override encoding format.
            codec_config (CodecConfig or None, optional): Codec configuration.
            filter_desc (str or None, optional): Additional filter processing.
            hw_accel (str or None, optional): Enable hardware acceleration.
        """
        self._stream_configs.append(("video", format))
        self._s.add_video_stream(
            frame_rate,
            width,
            height,
            format,
            encoder,
            encoder_option,
            encoder_format,
            encoder_frame_rate,
            encoder_width,
            encoder_height,
            hw_accel,
            _convert_config(codec_config),
            filter_desc,
        )

    def set_metadata(self, metadata: Dict[str, str]):
        """Set file-level metadata."""
        self._s.set_metadata(metadata)

    def open(self, option: Optional[Dict[str, str]] = None) -> "MediaEncoder":
        """Open the output file / device and write the header.

        Returns self for use as a context manager.
        """
        if not self._is_open:
            self._s.open(option)
            self._is_open = True
        return self

    def close(self):
        """Close the output."""
        if self._is_open:
            self._s.close()
            self._is_open = False

    def write_audio_chunk(self, i: int, chunk: torch.Tensor, pts: Optional[float] = None):
        """Write audio data.

        Args:
            i (int): Stream index.
            chunk (Tensor): Waveform tensor. Shape: ``(frame, channel)``.
            pts (float or None, optional): Presentation timestamp override.

        Note:
            The tensor is transferred via DLPack protocol for efficient zero-copy exchange.
        """
        chunk = chunk.contiguous()
        capsule = chunk.__dlpack__()
        self._s.write_audio_chunk_dlpack(i, capsule, pts)

    def write_video_chunk(self, i: int, chunk: torch.Tensor, pts: Optional[float] = None):
        """Write video/image data.

        Args:
            i (int): Stream index.
            chunk (Tensor): Video tensor. Shape: ``(time, channel, height, width)``.
            pts (float or None, optional): Presentation timestamp override.

        Note:
            The tensor is transferred via DLPack protocol for efficient zero-copy exchange.
        """
        _, fmt = self._stream_configs[i]
        chunk = _prepare_video_tensor(chunk, fmt)
        capsule = chunk.__dlpack__()
        self._s.write_video_chunk_dlpack(i, capsule, pts)

    def flush(self):
        """Flush the frames from encoders and write to the destination."""
        self._s.flush()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.flush()
        self.close()



import subprocess
import os
import sys
from pathlib import Path

from setuptools import setup, Extension

# ---------------------------------------------------------------------------
# Discover FFmpeg
# ---------------------------------------------------------------------------

FFMPEG_LIBS = [
    "libavcodec",
    "libavformat",
    "libavfilter",
    "libavutil",
]

FFMPEG_LIB_NAMES = ["avcodec", "avformat", "avfilter", "avutil", "swresample", "swscale"]


def _pkg_config(flag: str, libs: list) -> list:
    """Query pkg-config for compiler/linker flags."""
    try:
        out = subprocess.check_output(
            ["pkg-config", flag] + libs, text=True
        ).strip()
        return out.split() if out else []
    except (FileNotFoundError, subprocess.CalledProcessError):
        return []


def get_ffmpeg_config():
    """
    Get FFmpeg configuration for building.

    Priority:
    1. HUMECODEC_FFMPEG_ROOT environment variable (for wheel builds)
    2. PKG_CONFIG_PATH environment (for vendored FFmpeg)
    3. pkg-config system lookup
    4. Conda/pip environment
    """
    include_dirs = []
    library_dirs = []
    libraries = list(FFMPEG_LIB_NAMES)
    extra_compile_args = []
    extra_link_args = []

    # Check for explicit FFmpeg root (used in wheel builds)
    ffmpeg_root = os.environ.get("HUMECODEC_FFMPEG_ROOT")
    if ffmpeg_root:
        ffmpeg_root = Path(ffmpeg_root)
        if ffmpeg_root.exists():
            include_dir = ffmpeg_root / "include"
            lib_dir = ffmpeg_root / "lib"
            if include_dir.exists():
                include_dirs.append(str(include_dir))
            if lib_dir.exists():
                library_dirs.append(str(lib_dir))
                # Add rpath for the bundled libraries
                extra_link_args.append(f"-Wl,-rpath,$ORIGIN/../humecodec.libs")
            print(f"Using FFmpeg from HUMECODEC_FFMPEG_ROOT: {ffmpeg_root}")
            return {
                "include_dirs": include_dirs,
                "extra_compile_args": extra_compile_args,
                "libraries": libraries,
                "library_dirs": library_dirs,
                "extra_link_args": extra_link_args,
            }

    # Try pkg-config
    cflags = _pkg_config("--cflags", FFMPEG_LIBS)
    ldflags = _pkg_config("--libs", FFMPEG_LIBS)

    if cflags or ldflags:
        for f in cflags:
            if f.startswith("-I"):
                include_dirs.append(f[2:])
            else:
                extra_compile_args.append(f)

        libraries = []  # Reset - we'll get them from pkg-config
        for f in ldflags:
            if f.startswith("-l"):
                libraries.append(f[2:])
            elif f.startswith("-L"):
                library_dirs.append(f[2:])
            else:
                extra_link_args.append(f)

        if libraries:
            print(f"Using FFmpeg from pkg-config")
            return {
                "include_dirs": include_dirs,
                "extra_compile_args": extra_compile_args,
                "libraries": libraries,
                "library_dirs": library_dirs,
                "extra_link_args": extra_link_args,
            }

    # Fallback: try conda/pip environment
    env_prefix = sys.prefix
    env_lib = Path(env_prefix) / "lib"
    env_include = Path(env_prefix) / "include"

    # Check for FFmpeg in the environment
    if (env_lib / "libavcodec.so").exists() or (env_lib / "libavcodec.dylib").exists():
        library_dirs.append(str(env_lib))
        if env_include.exists():
            include_dirs.append(str(env_include))
        print(f"Using FFmpeg from Python environment: {env_prefix}")
    else:
        print("WARNING: FFmpeg not found. Build may fail.")
        print("  Set HUMECODEC_FFMPEG_ROOT or install FFmpeg via conda/pip.")

    return {
        "include_dirs": include_dirs,
        "extra_compile_args": extra_compile_args,
        "libraries": libraries,
        "library_dirs": library_dirs,
        "extra_link_args": extra_link_args,
    }


# ---------------------------------------------------------------------------
# Discover CUDA (without torch)
# ---------------------------------------------------------------------------

def get_cuda_config():
    """Detect CUDA toolkit for optional GPU support.

    Returns dict with include_dirs, library_dirs, libraries, and cuda_available.
    Detection priority:
    1. CUDA_HOME / CUDA_PATH environment variable
    2. nvcc on PATH
    3. Common install locations (/usr/local/cuda)
    """
    cuda_home = os.environ.get("CUDA_HOME") or os.environ.get("CUDA_PATH")

    if not cuda_home:
        # Try to find nvcc
        try:
            nvcc_path = subprocess.check_output(
                ["which", "nvcc"], text=True, stderr=subprocess.DEVNULL
            ).strip()
            if nvcc_path:
                # nvcc is typically at <cuda_home>/bin/nvcc
                cuda_home = str(Path(nvcc_path).parent.parent)
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass

    if not cuda_home:
        # Check common locations
        for candidate in ["/usr/local/cuda", "/opt/cuda"]:
            if Path(candidate).exists():
                cuda_home = candidate
                break

    if not cuda_home or not Path(cuda_home).exists():
        return {
            "include_dirs": [],
            "library_dirs": [],
            "libraries": [],
            "cuda_available": False,
        }

    cuda_home = Path(cuda_home)
    include_dir = cuda_home / "include"
    lib_dir = cuda_home / "lib64"
    if not lib_dir.exists():
        lib_dir = cuda_home / "lib"

    # Verify cudart exists
    has_cudart = any(
        (lib_dir / name).exists()
        for name in ["libcudart.so", "libcudart.dylib", "cudart.lib"]
    )

    if not has_cudart:
        print(f"WARNING: CUDA found at {cuda_home} but cudart library not found")
        return {
            "include_dirs": [],
            "library_dirs": [],
            "libraries": [],
            "cuda_available": False,
        }

    print(f"Using CUDA from: {cuda_home}")
    return {
        "include_dirs": [str(include_dir)] if include_dir.exists() else [],
        "library_dirs": [str(lib_dir)] if lib_dir.exists() else [],
        "libraries": ["cudart"],
        "cuda_available": True,
    }


# ---------------------------------------------------------------------------
# Discover pybind11
# ---------------------------------------------------------------------------

def get_pybind11_config():
    """Get pybind11 include directory."""
    try:
        import pybind11
        return {"include_dirs": [pybind11.get_include()]}
    except ImportError:
        raise RuntimeError(
            "pybind11 is required to build humecodec. "
            "Please install pybind11 first: pip install pybind11"
        )


# ---------------------------------------------------------------------------
# Collect C++ sources
# ---------------------------------------------------------------------------

CSRC = Path(__file__).parent / "src" / "humecodec" / "csrc"


def collect_sources():
    """Collect all C++ source files."""
    setup_dir = Path(__file__).parent
    return sorted(str(p.relative_to(setup_dir)) for p in CSRC.rglob("*.cpp"))


# ---------------------------------------------------------------------------
# Build extension
# ---------------------------------------------------------------------------

ffmpeg_cfg = get_ffmpeg_config()
cuda_cfg = get_cuda_config()
pybind_cfg = get_pybind11_config()

# Combine all include directories
include_dirs = (
    [str(Path(__file__).parent / "src")]  # so "humecodec/csrc/..." includes work
    + ffmpeg_cfg["include_dirs"]
    + cuda_cfg["include_dirs"]
    + pybind_cfg["include_dirs"]
)

# Combine all library directories
library_dirs = ffmpeg_cfg["library_dirs"] + cuda_cfg["library_dirs"]

# Combine all libraries
libraries = ffmpeg_cfg["libraries"] + cuda_cfg["libraries"]

# Compile arguments (MSVC uses different flags from GCC/Clang)
if sys.platform == "win32":
    extra_compile_args = ["/std:c++17"] + ffmpeg_cfg["extra_compile_args"]
else:
    extra_compile_args = ["-std=c++17", "-fPIC"] + ffmpeg_cfg["extra_compile_args"]

# Link arguments
extra_link_args = ffmpeg_cfg["extra_link_args"]

# Define macros
define_macros = []
if cuda_cfg["cuda_available"]:
    define_macros.append(("USE_CUDA", None))

ext = Extension(
    name="humecodec._humecodec",
    sources=collect_sources(),
    include_dirs=include_dirs,
    libraries=libraries,
    library_dirs=library_dirs,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    define_macros=define_macros,
    language="c++",
)

setup(
    ext_modules=[ext],
    options={"build_ext": {"parallel": True}},
)



_ext = None


def _load_ext():
    global _ext
    if _ext is None:
        from humecodec import _humecodec

        _ext = _humecodec
        _ext.init()
        _ext.set_log_level(16)  # AV_LOG_ERROR — suppress noisy warnings from decoders
    return _ext


def get_ext():
    return _load_ext()
